# Using the Factory Classes {#tutorial_factory}
@remark You only need to use the factories when you quickly want to create geometry which **is already** existing in the aircraft XML file.
If you start geometry from scratch you just need to use the geom2::MultisectionSurface and the geom2::SectionBuilder itself.
The surface provides all the necessary tools to build your geometry.

The factory classes follow the *factory design pattern*, hence the name.
The idea is, that you give this "factory" a **build plan** of what you want and the factory produces or creates the **item** for you.
In our case, the build plan is the *aircraft XML file* and the items are the *resulting surfaces*.

@note The concept how the geometry is encoded in den aircraft XML file **differs** from the concept used in this library!
The main difference is, that the library uses sections rather than segments!
The factories translate the data of the aircraft XML to the surface concept of this library.
Although, this might change in future release if the aircraft XML structure.

The library has factories for all major components of the aircraft geometry as they are currently defined in the aircraft XML file:
- @ref hull_factory
- @ref fuselage_factory
- @ref airfoil_surface_factory
- @ref wing_factory
- @ref spar_factory
- @ref control_device_factory

@note The Python examples assume that you imported the following modules:
```python
import pyaircraftGeometry2 as geom2
import pyaixml as aixml
```

---

## Preface
Before you can get started extracting the surfaces from the aircraft XML file, you need to have the file loaded as a `node` object/pointer.
It is best if you use the `std::filesystem::path` class to specify the file path.

- Load the aircraft XML file:

<div class="tabbed">
 
- <span class="tab-title">C++</span> 
```cpp
#include <filesystem>
#include <aixml/node.h>
std::filesystem::path file{"path/to/your/aircraft.xml"};
std::shared_ptr<node> AcXML = aixml::openDocument(file);
```
Since a `std::filesystem::path` can be constructed from a string, this is also possible:
```cpp
#include <aixml/node.h>
std::shared_ptr<node> AcXML = aixml::openDocument("path/to/your/aircraft.xml");
```
- <span class="tab-title">Python</span>
```python
AcXML = aixml.openDocument("path/to/your/aircraft.xml")
```

</div>

@attention Please always use `/` as the separator! The `std::filesystem` library takes care of converting this separator to the OS appropriate one.
However, it does not deal very well with strings where you use `\` or `\\` as the separator, since these are valid characters for file names on *UNIX* systems.

**Please note:**

For the sake of maintainability of this documentation, the following examples do not provide the content of the aircraft XML file. Instead, each example lists the XML file which is used for *unit testing* the class, since those files directly set the requirements of the XML content and are always up to date.

Due to changes in the format of the aircraft exchange file, there are subtle differences when using the factories with the newer version.
You can jump to @ref tutorial_aixml_v3 to directly get started with the newer version.
The factories support both versions and automatically deduce from the input data which version of the aircraft XML you are using.

---

## Aircraft XML Version 2.0.0 {#tutorial_aixml_v2}

@attention The factories **always** assume a relative path to `"AcftExchangeFile/Geometry"` when specifying paths to geometry nodes!
At least for now...

Most of the surfaces defined in the aircraft XML **version 2.0.0** file use `*.dat` files to specify the shape of the surface sections.
The paths to those files are denoted as *relative* paths.
The factories automatically deduce the *parent* path of the proved aircraft XML file and search for the geometry data **relative** to this parent path.

To illustrate this, let's assume you have the following directory structure:
```
/some/path/to/Project_Folder
├── aircraft.xml
└── geometryData
    ├── airfoilData
    │   ├── F15_11.dat
    │   └── n0012.dat
    ├── fuselage.dat
    └── nacelle.dat
```

By passing the node of the `aircraft.xml` file, the factories automatically deduce the parent path to be `/some/path/to/Project_Folder/` and remember this path internally.
It is then enough to specify the path, to i.e. the `n0012.dat` file, by adding this to the content of the aircraft XML file:
```xml
<SegmentPointData ToolLevel="1">geometryData/airfoilData/n0012.dat</SegmentPointData>
```

If the specified data file does not exist or the path is wrong, the library does **not** throw an error in the current implementation! When loading a `*.dat` file which does not exist, the library just returns a shape which contains a single point with the coordinates `[0.0, 0.0, 0.0]`. See @ref geom2.io.read_dat_file for more information.

### Hull Factory {#hull_factory}
> **XML Example:** `aircraftGeometry2/test/stubs/aixml-v2/hull.xml`

The hull factory is the most basic factory of the library and can be used to create simple hulls or tube surfaces, for example **nacelles**.
All described functionality of this factory does also apply to all other factories since they are all derived from the same base factory class.

- Create a hull factory:

<div class="tabbed">

- <span class="tab-title">C++</span> 
```cpp
/* Create the hull factory */
#include <aircraftGeometry2/hull_surface.h>
geom2::HullFactory factory{AcXML, "./geometryData"};
```
- <span class="tab-title">Python</span>
```python
# Create the hull factory
factory = geom2.factory.HullFactory(AcXML, "./geometryData")
```

</div>

@note You can omit the path `"./geometryData"` for the *version 2.0.0* aircraft exchange files and just give an empty string. The factories then will search for the directory as explained before.

After creating the factory with the existing aircraft XML data, the factory knows what to do and you can ask it to create the complete surface for you.
The aircraft XML can contain several surfaces of the same type, but with a different id.
You can have multiple *nacelles* for example.
That is why you have to tell the factory which surface you want to extract from the aircraft XML.
You can do this by specifying the **surface name** and its **id**.
You can use the id notation as provided by the **aixml** library.

You are not limited to just creating *nacelles* with this factory.
As long as the data structure of the custom surface follows the convention as defined in `aircraftGeometry2/test/stubs/acxml-v2/hull.xml` it can be created with this factory.

- Get **nacelle** with the id `SimpleHull` from the factory:

<div class="tabbed">
 
- <span class="tab-title">C++</span> 
```cpp
/* Create the nacelle surface */
geom2::MultisectionSurface<geom2::PolygonSection> nacelle = factory.create("Nacelle@SimpleHull");
```
- <span class="tab-title">Python</span>
```python
# Create the nacelle surface
nacelle = factory.create("Nacelle@SimpleHull")
```

</div>

@attention The current aircraft XML does not define the orientation of those surfaces, yet.
Therefore, the normal direction when retrieving the surface from the factory is always the default direction.
The origin point, however, is extracted from the aircraft XML file.

The resulting multi-section surface is **moved** from the factory!
This means you can use the same factory multiple times and create different surfaces by passing different surface id's.

If you only need the factory **once**, you can abbreviate the process as follows:

<div class="tabbed">
 
- <span class="tab-title">C++</span> 
```cpp
/* Create the nacelle surface */
geom2::MultisectionSurface<geom2::PolygonSection> nacelle = geom2::HullFactory{AcXML, ""}.create("Nacelle@SimpleHull");
```
- <span class="tab-title">Python</span>
```python
# Create the nacelle surface
nacelle = geom2.factory.HullFactory(AcXML, "").create("Nacelle@SimpleHull")
```

</div>

### Fuselage Factory {#fuselage_factory}
> **XML Example:** `aircraftGeometry2/test/stubs/aixml-v2/fuselage.xml`

The fuselage factory does basically the same thing as the hull factory.
But since the aircraft XML currently defines the data structure for fuselages a bit differently than for nacelles, this specialized factory exists.

- Create the fuselage surface:

<div class="tabbed">

- <span class="tab-title">C++</span> 
```cpp
/* Create the factory */
#include <aircraftGeometry2/fuselage.h>
geom2::FuselageFactory factory{AcXML, ""};

/* Create the surface */
geom2::MultisectionSurface<geom2::PolygonSection> fuselage = factory.create("Fuselage@SimpleFuselage");
```
- <span class="tab-title">Python</span>
```python
# Create the factory
factory = geom2.factory.FuselageFactory(AcXML, "")

# Create the surface
fuselage = factory.create("Fuselage@SimpleFuselage")
```

</div>

### Airfoil Surface Factory {#airfoil_surface_factory}
> **XML Example:** `aircraftGeometry2/test/stubs/aixml-v2/pylon.xml`

The airfoil surface factory can be used to create the following surfaces from the aircraft XML file:
- **Pylon**
- **VerticalSurface**

The main difference of this factory is, that is uses geom2::AirfoilSection when creating the multi-section surface.

- Create a pylon surface:

<div class="tabbed">

- <span class="tab-title">C++</span> 
```cpp
/* Create the factory */
#include <aircraftGeometry2/airfoil_surface.h>
geom2::AirfoilSurfaceFactory factory{AcXML, ""};

/* Create the surface */
geom2::MultisectionSurface<geom2::AirfoilSection> pylon = factory.create("Pylon@SimplePylon");
```
- <span class="tab-title">Python</span>
```python
# Create the factory
factory = geom2.factory.AirfoilSurfaceFactory(AcXML, "")

# Create the surface
pylon = factory.create("Pylon@SimplePylon")
```

</div>

@attention As recommended in @ref tutorial_airfoil_surface , the airfoil factories use the **negative** local `Z` direction as the extrusion or span direction!

### Wing Factory {#wing_factory}
> **XML Example:** `aircraftGeometry2/test/stubs/aixml-v2/wing.xml`

The wing factory specializes the airfoil surface factory.
Again, due to a slight different data structure in the aircraft XML file.
The functionality is otherwise the same.

- Create a wing surface:

<div class="tabbed">

- <span class="tab-title">C++</span> 
```cpp
/* Create the factory */
#include <aircraftGeometry2/airfoil_surface.h>
geom2::WingFactory factory{AcXML, ""};

/* Create the surface */
geom2::MultisectionSurface<geom2::AirfoilSection> wing = factory.create("LiftingSurface@MainWing");
```
- <span class="tab-title">Python</span>
```python
# Create the factory
factory = geom2.factory.WingFactory(AcXML, "")

# Create the surface
wing = factory.create("LiftingSurface@MainWing")
```

</div>

@note The wing factory does **not** produces discontinuous sections!
It takes the inner length of the first segment as the length for the first section and then continuous by using the outer length of the following segments from the aircraft XML file!

### Spar Factory {#spar_factory}
> **XML Example:** `aircraftGeometry2/test/stubs/aixml-v2/wing.xml`

The aircraft XML file provides the relative positions of the front and rear spar location of the lifting surfaces.
The spar factory creates a planar surface using those relative coordinates.
Each section of the spar contains two coordinates:
The **front** and the **rear** position in relative coordinates ranging from *0* to *1*.
The position is defined by the local `X` direction, the `Y` coordinate of the section points is **always** *0.0*!
The `Z` offset of these sections corresponds to the relative span position and also ranges from *0* to *1*.

- Create the spar of a wing:

<div class="tabbed">

- <span class="tab-title">C++</span> 
```cpp
/* Create the factory */
#include <aircraftGeometry2/airfoil_surface.h>
geom2::SparFactory factory{AcXML, ""};

/* Create the surface */
geom2::MultisectionSurface<geom2::PolygonSection> spar = factory.create("LiftingSurface@MainWing");
```
- <span class="tab-title">Python</span>
```python
# Create the factory
factory = geom2.factory.SparFactory(AcXML, "")

# Create the surface
spar = factory.create("LiftingSurface@MainWing")
```

</div>

@note This factory returns a multi-section surface which contains geom2::PolygonSection as the section type.
Be aware, that the spar geometry is **not** considered an airfoil shape in the context of this library.
The extrusion direction, however, is the **negative** local `Z` direction as for the airfoil surfaces, so that the spar geometry is at the same location as the wing geometry.
The origin point of the spar is the same as the wing origin as well.

The spar factory just returns the spar surface and **not** the wing surface.
You have to use the wing factory separately to create the wing surface.

### Control Device Factory {#control_device_factory}
> **XML Example:** `aircraftGeometry2/test/stubs/aixml-v2/wing.xml`

As for the spar geometry, the control device geometry is coupled to the wing geometry it belongs to.
The handling of control devices is different from the other geometry objects in the way, that all dimensions are normalized between [0, 1].
The local `Z` coordinates are normalized using the **span** of the corresponding wing and the local `X` coordinates are normalized using the **chord length** of the wing at the corresponding location of the control device.
Since the control devices are modelled as "flat plates", they do **not** have a *thickness* or local `Y` coordinates greater than 0.

You can use the control device factory to extract single devices from the aircraft XML file using geom2::ControlDeviceFactory::create.

- Create the geometry of a control device of a wing:

<div class="tabbed">

- <span class="tab-title">C++</span> 
```cpp
/* Create the factory */
#include <aircraftGeometry2/airfoil_surface.h>
geom2::ControlDeviceFactory factory{AcXML, ""};

/* Create the surface */
geom2::MultisectionSurface<geom2::PolygonSection> flap = factory.create("LiftingSurface@MainWing/SurfaceParameters/HalfSurfaceDescription/ControlDeviceSetup/TEDevice@1");
```
- <span class="tab-title">Python</span>
```python
# Create the factory
factory = geom2.factory.ControlDeviceFactory(AcXML, "")

# Create the surface
flaps = factory.create("LiftingSurface@MainWing/SurfaceParameters/HalfSurfaceDescription/ControlDeviceSetup/TEDevice@1")
```

</div>

---

## Aircraft XML Version 3.0.0 {#tutorial_aixml_v3}

@attention The factories **always** assume a relative path to `"aircraft_exchange_file/component_design/"` when specifying paths to geometry nodes!
At least for now...

The newer version of the aircraft XML file does not expect the geometry files at a certain location as a result of the modularization of the whole project.
Instead, each surface factory of this library expects a path where to find the geometry files.
So when using the factories, you **must** specify this *data directory* otherwise the factories will not know where to look for the geometry.
The node defining the geometry file in the aircraft XML simply looks like this:
```xml
<value>n0012</value>
```
The library will then try to use a `n0012.dat` file from the `data_directory` you have given it.

@note In the following examples it is assumed, that you have the same directory structure for the *version 3.0.0* XML files as for the *version 2.0.0* files in @ref tutorial_aixml_v2 !

### Hull Factory 
> **XML Example:** `aircraftGeometry2/test/stubs/aixml-v3/hull.xml`

The hull factory is the most basic factory of the library and can be used to create simple hulls or tube surfaces, for example **nacelles**.
All described functionality of this factory does also apply to all other factories since they are all derived from the same base factory class.

- Create a hull factory:

<div class="tabbed">

- <span class="tab-title">C++</span> 
```cpp
/* Create the hull factory */
#include <aircraftGeometry2/hull_surface.h>
geom2::HullFactory factory{AcXML, "./geometryData"};
```
- <span class="tab-title">Python</span>
```python
# Create the hull factory
factory = geom2.factory.HullFactory(AcXML, "./geometryData")
```

</div>

@note You **cannot** omit the path `"./geometryData"` for the *version 3.0.0* aircraft exchange files!

- Get **nacelle** with the id `0` from the factory:

<div class="tabbed">
 
- <span class="tab-title">C++</span> 
```cpp
/* Create the nacelle surface */
geom2::MultisectionSurface<geom2::PolygonSection> nacelle = factory.create("propulsion/specific/geometry/nacelle@0");
```
- <span class="tab-title">Python</span>
```python
# Create the nacelle surface
nacelle = factory.create("propulsion/specific/geometry/nacelle@0")
```

</div>

Note that the path to the nacelle node needs more information than in @ref tutorial_aixml_v2 since the structure of the aircraft exchange file *version 3.0.0* allows for a more flexible arrangement for different components.
Since those components still use the some geometry encoding throughout the file, the same factories can be used to create the geometry surfaces.
However, the factories cannot know where to find each individual component.
That is why you have to give the more detailed node path when creating the surfaces, so that the factory knows where to find the specific nodes.
This makes the factory more flexible, since you are not confined to certain locations within the exchange file to extract geometry from, but you can use any node from the exchange file as long as its content structure satisfies the requirements of the surface. (Remember: The requirements are documented in the given XML example files.)

### Fuselage Factory 
> **XML Example:** `aircraftGeometry2/test/stubs/aixml-v3/fuselage.xml`

The fuselage factory does basically the same thing as the hull factory.
But since the aircraft XML currently defines the data structure for fuselages a bit differently than for nacelles, this specialized factory exists.

- Create the fuselage surface:

</div>
`tbd`

### Airfoil Surface Factory 
> **XML Example:** `aircraftGeometry2/test/stubs/aixml-v3/pylon.xml`

`tbd`

@attention As recommended in @ref tutorial_airfoil_surface , the airfoil factories use the **negative** local `Z` direction as the extrusion or span direction!

### Wing Factory 
> **XML Example:** `aircraftGeometry2/test/stubs/aixml-v3/wing.xml`

The wing factory creates the aerodynamic surfaces of the wings.

- Create a wing surface:

<div class="tabbed">

- <span class="tab-title">C++</span> 
```cpp
/* Create the factory */
#include <aircraftGeometry2/airfoil_surface.h>
geom2::WingFactory factory{AcXML, "./geometryData/airfoilData"};

/* Create the surface */
geom2::MultisectionSurface<geom2::AirfoilSection> wing = factory.create("wing/specific/geometry/aerodynamic_surface@0");
```
- <span class="tab-title">Python</span>
```python
# Create the factory
factory = geom2.factory.WingFactory(AcXML, "./geometryData/airfoilData")

# Create the surface
wing = factory.create("wing/specific/geometry/aerodynamic_surface@0")
```

</div>

### Spar Factory 
> **XML Example:** `aircraftGeometry2/test/stubs/aixml-v3/wing.xml`

The aircraft XML file provides the relative positions of the front and rear spar location of the lifting surfaces.
The spar factory creates a planar surface using those relative coordinates.
Each section of the spar contains two coordinates:
The **front** and the **rear** position in relative coordinates ranging from *0* to *1*.
The position is defined by the local `X` direction, the `Y` coordinate of the section points is **always** *0.0*!
The `Z` offset of these sections corresponds to the relative span position and also ranges from *0* to *1*.

- Create the spar of a wing:

<div class="tabbed">

- <span class="tab-title">C++</span> 
```cpp
/* Create the factory */
#include <aircraftGeometry2/airfoil_surface.h>
geom2::SparFactory factory{AcXML, "./geometryData/airfoilData"};

/* Create the surface */
geom2::MultisectionSurface<geom2::PolygonSection> spar = factory.create("wing/specific/geometry/aerodynamic_surface@0/spars/spar@0");
```
- <span class="tab-title">Python</span>
```python
# Create the factory
factory = geom2.factory.SparFactory(AcXML, "./geometryData/airfoilData")

# Create the surface
spar = factory.create("wing/specific/geometry/aerodynamic_surface@0/spars/spar@0")
```

</div>

@note This factory returns a multi-section surface which contains geom2::PolygonSection as the section type.
Be aware, that the spar geometry is **not** considered an airfoil shape in the context of this library.
The extrusion direction, however, is the **negative** local `Z` direction as for the airfoil surfaces, so that the spar geometry is at the same location as the wing geometry.
The origin point of the spar is the same as the wing origin as well.

The spar factory just returns the spar surface and **not** the wing surface.
You have to use the wing factory separately to create the wing surface.

### Control Device Factory 
> **XML Example:** `aircraftGeometry2/test/stubs/aixml-v3/wing.xml`

As for the spar geometry, the control device geometry is coupled to the wing geometry it belongs to.
The handling of control devices is different from the other geometry objects in the way, that all dimensions are normalized between [0, 1].
The local `Z` coordinates are normalized using the **span** of the corresponding wing and the local `X` coordinates are normalized using the **chord length** of the wing at the corresponding location of the control device.
Since the control devices are modelled as "flat plates", they do **not** have a *thickness* or local `Y` coordinates greater than 0.

You can use the control device factory to extract single devices from the aircraft XML file using geom2::ControlDeviceFactory::create.

- Create the geometry of a control device of a wing:

<div class="tabbed">

- <span class="tab-title">C++</span> 
```cpp
/* Create the factory */
#include <aircraftGeometry2/airfoil_surface.h>
geom2::ControlDeviceFactory factory{AcXML, "./geometryData/airfoilData"};

/* Create the surface */
geom2::MultisectionSurface<geom2::PolygonSection> flap = factory.create("wing/specific/geometry/aerodynamic_surface@0/control_devices/control_device@0");
```
- <span class="tab-title">Python</span>
```python
# Create the factory
factory = geom2.factory.ControlDeviceFactory(AcXML, "./geometryData/airfoilData")

# Create the surface
flaps = factory.create("wing/specific/geometry/aerodynamic_surface@0/control_devices/control_device@0")
```

</div>

---
