/*  Copyright (C) 2010-2023 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "blackboxTest/blackboxTest.h"
#include <runtimeInfo/runtimeInfo.h>
#include <standardFiles/functions.h>
#include <standardFiles/settings.h>
#include <vector>
#include <limits>

/* Name convention */
// using BlackboxDeathTest = BlackboxTest;
// using BlackboxInducedDeathTest = BlackboxTest;

/*
 * Defines to *temporarily* configure this library!
 */
// #define ENABLE_TOOL_LEVELS
// #define ENABLE_INDUCED_DEATH_TESTS

/* Initialization of static variables */
node* BlackboxTest::blackboxTestData{nullptr};
int8_t BlackboxTest::currentAssertion{-1};

namespace detail {
/**
 * @brief Visitor class to convert the any value to a string.
 */
struct StringVisitor {
  /* Overloaded operator() to convert the any value to a string */
  std::string operator()(double value) { return std::to_string(value); }
  std::string operator()(const std::string_view& value) { return std::string{value}; }
  std::string operator()(const std::string& value) { return value; }
};  // struct AnyGet
}  // namespace detail

std::unordered_map<std::string, BlackboxTest::TestCaseData>& BlackboxTest::getTestCaseMap() {
    //Static map inside function to resolve static initialization problems
    //see: https://isocpp.org/wiki/faq/ctors#static-init-order-on-first-use-members
    static std::unordered_map<std::string, TestCaseData> * res = new std::unordered_map<std::string, TestCaseData>();
    return *res;
}

std::unordered_map<std::string, std::string>& BlackboxTest::getTestParamPathsMap() {
    //Static map inside function to resolve static initialization problems
    //see: https://isocpp.org/wiki/faq/ctors#static-init-order-on-first-use-members
    static std::unordered_map<std::string, std::string> * res = new std::unordered_map<std::string, std::string>();
    return *res;
}

struct BlackboxTest::TestCaseData {
    node* testCaseNode{}; /**< Node pointer to a node containing test case information of the respective test */
    bool executability{}; /**< Flag informing whether a test case is executable or not. If not, later the test will be skipped. */

    /** \brief Default constructor for TestCaseData
     */
    TestCaseData() = default;

    /** \brief Constructor for TestCaseData
     * \param testCaseNode: Node pointer to a node containing test case information of the respective test
     * \param executability: Flag informing whether a test case is executable or not. If not, later the test will be skipped.
     */
    TestCaseData(node* testCaseNode, bool executability) : testCaseNode(testCaseNode), executability(executability) {}
};

void BlackboxTest::Environment::TearDown() { //cppcheck-suppress unusedFunction
    aixml::closeDocument(*BlackboxTest::blackboxTestData);
    delete &BlackboxTest::getTestParamPathsMap();
}

void BlackboxTest::registerTests(const std::string& tool_name) {
    std::vector<node*> testSuites = blackboxTestData->getVector("/BlackboxTestSuites/TestSuite");
    for (node* currTestSuite : testSuites) {
        const std::string TestSuiteName =  currTestSuite->getStringAttrib("Name");
        std::vector<node*> tests = currTestSuite->getVector("/Test");

        for (node* currTest : tests) {
            const std::string TestName = currTest->getStringAttrib("Name");
            std::vector<node*> testCases = currTest->getVector("/TestCase");
            BlackboxTest::addTestPathToMaps(TestSuiteName + "." + TestName, currTest);

            for (node* currTestCase : testCases) {
            //Test Suite Name must correspond to fixture class name!

                const std::string TestCaseValue = currTestCase->at("/TestCaseValue");

                const std::string TestNameWithValue = TestName + "(" + TestCaseValue + ")";
                testing::RegisterTest(TestSuiteName.c_str(), TestNameWithValue.c_str(), nullptr, nullptr,
                                        __FILE__, __LINE__,
                                        // Important to use the fixture type as the return type here.
                                        [=]() -> BlackboxTest* { return new BlackboxTest(tool_name); });

                BlackboxTest::addCaseToMaps(TestSuiteName + "." + TestNameWithValue, currTestCase);
#ifdef ENABLE_INDUCED_DEATH_TESTS
                // If Test is not already a Death Test register an additional Death Test to test for premature exit
                // All TestSuite names of Death Tests corresponding to exit-tests end in "InducedDeathTest"
                if (TestSuiteName.find("Death") == std::string::npos) {
                    const std::string InducedTestSuiteName = TestSuiteName + "InducedDeathTest";
                    testing::RegisterTest(InducedTestSuiteName.c_str(), TestNameWithValue.c_str(), nullptr, nullptr,
                                        __FILE__, __LINE__,
                                        // Important to use the fixture type as the return type here.
                                        [=]() -> BlackboxTest* { return new BlackboxTest(tool_name); });

                    BlackboxTest::addTestPathToMaps(InducedTestSuiteName + "." + TestName, currTest);
                    BlackboxTest::addCaseToMaps(InducedTestSuiteName + "." + TestNameWithValue, currTestCase);
                }
#endif
            }
        }
    }
}

void BlackboxTest::addCaseToMaps(std::string name, node* testCaseNode) {
#ifdef ENABLE_INDUCED_DEATH_TESTS
    getTestCaseMap()[name] = TestCaseData(testCaseNode, false);
#else
    getTestCaseMap()[name] = TestCaseData(testCaseNode, true);
#endif
}

void BlackboxTest::addTestPathToMaps(std::string name, node* testNode) {
    std::string path = testNode->at("/TestParamPath");
    getTestParamPathsMap()[name] = path;
}

BlackboxTest::BlackboxTest(const std::string& tool_name)
    : tool_name(tool_name)
    , test_acXML(nullptr)
    , configXML(aixml::openDocument(getSettingsFilename(tool_name))) {
    /* Get name of test and testSuite currently running */
    const testing::TestInfo* const test_info = testing::UnitTest::GetInstance()->current_test_info(); //don't delete pointer! gtest takes care of it
    currentTestCaseName = test_info->name();
    currentTestSuiteName = test_info->test_suite_name();
}

void BlackboxTest::SetupTest() {
    if (myRuntimeInfo == nullptr) {
        myRuntimeInfo = new runtimeInfo(checkBoundaries(this->configXML.at("ConsoleOutputOn"), 0, true, 3, true),
                                        checkBoundaries(this->configXML.at("LogfileOutputOn"), 0, true, 3, true),
                                        this->configXML.at("LogFile"), this->tool_name);
    }
}

void BlackboxTest::TeardownTest() {
    /* Reset altered aircraft xml parameter back to original value  */
    if (!aircraftXmlParameterPair.first.empty()) {
        const settings mySettings(this->configXML);
        node& acXML(aixml::openDocument(mySettings.getIoDir() + mySettings.getIoFilename())); // Open aircraft xml file
        if (isNumber(aircraftXmlParameterPair.second)) {
            acXML.at(aircraftXmlParameterPair.first) = std::stod(this->preTestValue); // Reset altered aircraft xml parameter to original value
        } else {
            acXML.at(aircraftXmlParameterPair.first) = this->preTestValue; // Reset altered aircraft xml parameter to original value
        }
#ifdef ENABLE_TOOL_LEVELS
        if (acXML.at(aircraftXmlParameterPair.first).hasAttrib("ToolLevel")) {
            if (acXML.at(aircraftXmlParameterPair.first).getIntAttrib("ToolLevel") > mySettings.getToolLevel()) {
                aixml::saveDocument(acXML, acXML.at(aircraftXmlParameterPair.first).getIntAttrib("ToolLevel"));
            } else {
                aixml::saveDocument(acXML, mySettings.getToolLevel());
            }
        } else {
            aixml::saveDocument(acXML, mySettings.getToolLevel());
        }
#endif
        aixml::closeDocument(acXML);
        this->aircraftXmlParameterPair.first.clear(); // Clear to be altered pair of aircraft xml parameter + value
        this->aircraftXmlParameterPair.second.clear();
        this->preTestValue.clear();
    }
    if (myRuntimeInfo) {
        delete myRuntimeInfo;
        myRuntimeInfo = nullptr;
    }
}

std::string BlackboxTest::generateErrorMessage(int8_t assertType, const std::string &testParamPath,
                                                                    const std::string &testParamValue, const std::string &criticalParamPath, parameter_value criticalParamValue,
                                                                    const std::string &assertionValue, double acc) {
    bool outputType = !human_readable_output;
    std::string message = "&&&";
    if (!outputType) {
        message += getHumanReadableErrorMessage(assertType, testParamPath, testParamValue,
                                                criticalParamPath, criticalParamValue, assertionValue, acc);
    } else {
        message += getMachineReadableErrorMessage();
    }
    return message;
}

std::string BlackboxTest::generateErrorMessage(bool inducedDeathTest, const std::string &testParamPath, const std::string &testParamValue) {
    bool outputType = !human_readable_output;
    std::string message = "&&&";
    if (!outputType) {
        message += getHumanReadableErrorMessage(inducedDeathTest, testParamPath, testParamValue);
    } else {
        message += getMachineReadableErrorMessage();
    }
    return message;
}

std::string BlackboxTest::getHumanReadableErrorMessage(int8_t assertType, const std::string &testParamPath,
                                                                   const std::string &testParamValue, const std::string &criticalParamPath, parameter_value criticalParamValue,
                                                                   const std::string &assertionValue, double acc) {
    std::string message = "\nTEST CASE FAILED\n";
    message += "Tested Parameter:\t" + testParamPath + "\n";
    message += "Tested Value:\t\t" + testParamValue + "\n";
    message += "Critical Parameter:\t" + criticalParamPath + "\n";

    /* Convert either value of the parameter variant to a string */
    std::string value_string = std::visit(detail::StringVisitor{}, criticalParamValue);

    switch (assertType) {
                case STREQ: {
                    message += "Expected:\t\t" + assertionValue + "\nResult:\t\t\t" + value_string + "\n";
                    break;
                }
                case EQ: {
                    message += "Expected:\t\t" + value_string + " == " + assertionValue + "\n";
                    break;
                }
                case LE: {
                    message += "Expected:\t\t" + value_string + " <= " + assertionValue + "\n";
                    break;
                }
                case GE: {
                    message += "Expected:\t\t" + value_string + " >= " + assertionValue + "\n";
                    break;
                }
                case LT: {
                    message += "Expected:\t\t " + value_string + " < " + assertionValue + "\n";
                    break;
                }
                case GT: {
                    message += "Expected:\t\t" + value_string + " > " + assertionValue + "\n";
                    break;
                }
                case NEAR: {
                    message += "Expected:\t\t|" + value_string + " - " + assertionValue + "| < " + std::to_string(acc) + "\n";
                    break;
                }
                default:
                    message += "Unknown expectation\n";
            }
    return message;
}

std::string BlackboxTest::getHumanReadableErrorMessage(bool inducedDeathTest, const std::string &testParamPath, const std::string &testParamValue) {
    std::string message = "\nDEATH TEST CASE FAILED\n";
    if (inducedDeathTest) {
        message += "Module calculation exited prematurely. Corresponding Test Case is skipped.\n";
    } else {
        message += "Module calculation failed to die.\n";
    }
    message += "Tested Parameter:\t" + testParamPath + "\n";
    message += "Tested Value:\t\t" + testParamValue + "\n";
    return message;
}

std::string BlackboxTest::getMachineReadableErrorMessage() {
    std::string message("\nTEST FAILED");
    return message;
}

int8_t BlackboxTest::getAssertion(const std::string& readValue) {
    currentAssertion = -1;
    if (!readValue.compare("LE") || !readValue.compare("le")) {
        currentAssertion = LE;
    } else if (!readValue.compare("EQ") || !readValue.compare("eq")) {
        currentAssertion = EQ;
    } else if (!readValue.compare("STREQ") || !readValue.compare("streq")) {
        currentAssertion = STREQ;
    } else if (!readValue.compare("GE") || !readValue.compare("ge")) {
        currentAssertion = GE;
    } else if (!readValue.compare("LT") || !readValue.compare("lt")) {
        currentAssertion = LT;
    } else if (!readValue.compare("GT") || !readValue.compare("gt")) {
        currentAssertion = GT;
    } else if (!readValue.compare("NEAR") || !readValue.compare("near")) {
        currentAssertion = NEAR;
    }
    return currentAssertion;
}

void BlackboxTest::TestBody() { //cppcheck-suppress unusedFunction
    // Check if the test currently running is inside a Death Test Suite
    if (currentTestSuiteName.find("Death") != std::string::npos) {
        try {
            standardDeathTest();
        } catch (...) {
            std::cout << "Uncaught failure in death test" << std::endl;
        }
    } else {
        standardTest();
    }
}

void BlackboxTest::applyAssertions(const node& criticalParamsNode, node* currentTestCaseNode, const std::string &testParamPath) {
    std::vector<node*> critParams = criticalParamsNode.getVector("/CriticalParameter");
    /* Loop over all assertions for all defined critical parameters for the current test case */
    for (node* currCriticalParameter : critParams) {
        std::string currCritParamPath = currCriticalParameter->at("/CriticalParamPath");
        std::vector<node*> assertions = currCriticalParameter->getVector("/Assertions/Assertion");
        for (node* currAssertion : assertions) {
            std::string type = currAssertion->getStringAttrib("Type");

            int8_t typeCode = BlackboxTest::getAssertion(type);
            std::string testCaseValue = currentTestCaseNode->at("/TestCaseValue");
            std::string message("");

            /* Check whether assertionValue is a valid XML path */
            // -> Skip STREQ assertion in case the comparison string is a valid XML path
            node* assertionNode = nullptr;
            if (typeCode != STREQ) {
                assertionNode = this->test_acXML->find(*currAssertion);
            }
            std::string assertionValue = assertionNode ? *assertionNode : *currAssertion;

            /* Apply corresponding assertions */
            switch (typeCode) {
                case EQ: {
                    double currCritParamValue = this->test_acXML->at(currCritParamPath);
                    //std::cout << "EXPCECT_EQ(" << currCritParamValue << "," << assertionValue << ")" << std::endl;
                    message = generateErrorMessage(typeCode, testParamPath, testCaseValue , currCritParamPath, currCritParamValue, assertionValue);
                    EXPECT_EQ(currCritParamValue, std::stod(assertionValue)) << message;
                    break;
                }
                case STREQ: {
                    std::string currCritParamValue = this->test_acXML->at(currCritParamPath);
                    //std::cout << "EXPCECT_EQ(" << currCritParamValue << "," << assertionValue << ")" << std::endl;
                    message = generateErrorMessage(typeCode, testParamPath, testCaseValue , currCritParamPath, std::string_view{currCritParamValue}, assertionValue);
                    EXPECT_EQ(currCritParamValue, assertionValue) << message;
                    break;
                }
                case LE: {
                    double currCritParamValue = this->test_acXML->at(currCritParamPath);
                    //std::cout << "EXPCECT_LE(" << currCritParamValue << "," << assertionValue << ")" << std::endl;
                    message = generateErrorMessage(typeCode, testParamPath, testCaseValue , currCritParamPath, currCritParamValue, assertionValue);
                    EXPECT_LE(currCritParamValue, std::stod(assertionValue)) << message;
                    break;
                }
                case GE: {
                    double currCritParamValue = this->test_acXML->at(currCritParamPath);
                    //std::cout << "EXPCECT_GE(" << currCritParamValue << "," << assertionValue << ")" << std::endl;
                    message = generateErrorMessage(typeCode, testParamPath, testCaseValue , currCritParamPath, currCritParamValue, assertionValue);
                    EXPECT_GE(currCritParamValue, std::stod(assertionValue)) << message;
                    break;
                }
                case LT: {
                    double currCritParamValue = this->test_acXML->at(currCritParamPath);
                    //std::cout << "EXPCECT_LT(" << currCritParamValue << "," << assertionValue << ")" << std::endl;
                    message = generateErrorMessage(typeCode, testParamPath, testCaseValue , currCritParamPath, currCritParamValue, assertionValue);
                    EXPECT_LT(currCritParamValue, std::stod(assertionValue)) << message;
                    break;
                }
                case GT: {
                    double currCritParamValue = this->test_acXML->at(currCritParamPath);
                    //std::cout << "EXPCECT_GT(" << currCritParamValue << "," << assertionValue << ")" << std::endl;
                    message = generateErrorMessage(typeCode, testParamPath, testCaseValue , currCritParamPath, currCritParamValue, assertionValue);
                    EXPECT_GT(currCritParamValue, std::stod(assertionValue)) << message;
                    break;
                }
                case NEAR: {
                    double currCritParamValue = this->test_acXML->at(currCritParamPath);
                    //std::cout << "EXPCECT_NEAR(" << currCritParamValue << "," << assertionValue << ")" << std::endl;
                    double accuracy{0};
                    if (currAssertion->hasAttrib("Accuracy")) {
                        accuracy = currAssertion->getDoubleAttrib("Accuracy");
                    } else if (currAssertion->hasAttrib("relative")) {
                        accuracy = currAssertion->getDoubleAttrib("relative") * std::stod(assertionValue);
                    } else {
                        throw std::runtime_error("No accuracy specified for NEAR assertion.");
                    }
                    message = generateErrorMessage(typeCode, testParamPath, testCaseValue , currCritParamPath, currCritParamValue, assertionValue, accuracy);
                    EXPECT_NEAR(currCritParamValue, std::stod(assertionValue), accuracy) << message;
                    break;
                }
                default:
                    std::cout << "Unknown assertion is skipped." << std::endl;
            }

            if (HasFailure()) {
                RecordProperty("altmessage", message);
            }
        }
    }
}

void BlackboxTest::standardTest() {
    std::string completeName = currentTestSuiteName + "." + currentTestCaseName;
    std::string testParameterPath = getTestParamPathsMap()[completeName.substr(0, completeName.find("("))];
    if (!blackboxTestData->find("/BlackboxTestSuites/TestSuite@Name=" + currentTestSuiteName + "/CriticalParameters")) {
        std::cout << "No node /CriticalParameters found in current TestSuite. Please add node first." << std::endl;
        FAIL();
    }

    const node criticalParamsNode(blackboxTestData->at("/BlackboxTestSuites/TestSuite@Name=" + currentTestSuiteName + "/CriticalParameters"));

    /* Check whether current test should be skipped */
    // Remove the bracket part of the testcasename
    std::string testName = currentTestCaseName.substr(0, currentTestCaseName.find("("));
    const node currentTest = blackboxTestData->at("/BlackboxTestSuites/TestSuite@Name=" + currentTestSuiteName + "/Test@Name=" + testName);
    if (currentTest.hasAttrib("skip")) {
        GTEST_SKIP() << "  Skip reason: " << currentTest.getStringAttrib("skip");
    }

    //Phase 1: Setup
    SetupTest();

    //Phase 2: Alteration of test parameter
    node* currentTestCase = getTestCaseMap()[completeName].testCaseNode;
    std::string testParameterValue = currentTestCase->at("/TestCaseValue");
    try {
        if (testParameterPath.find("ConfigFile") != std::string::npos) {
            configXML.at(testParameterPath) = testParameterValue;
        } else if (testParameterPath.find("AcftExchangeFile") != std::string::npos) {
            aircraftXmlParameterPair = std::make_pair(testParameterPath, testParameterValue);
            const settings mySettings(this->configXML);
            node& acXML(aixml::openDocument(mySettings.getIoDir() + mySettings.getIoFilename()));
            if (acXML.find(aircraftXmlParameterPair.first)) {
                const std::string DUMMYpreTestValue = acXML.at(aircraftXmlParameterPair.first);
                this->preTestValue = DUMMYpreTestValue;
                if (isNumber(aircraftXmlParameterPair.second)) {
                    acXML.at(aircraftXmlParameterPair.first) = std::stod(aircraftXmlParameterPair.second);
                } else {
                    acXML.at(aircraftXmlParameterPair.first) = aircraftXmlParameterPair.second;
                }
#ifdef ENABLE_TOOL_LEVELS
                if (acXML.at(aircraftXmlParameterPair.first).hasAttrib("ToolLevel")) {
                    if (acXML.at(aircraftXmlParameterPair.first).getIntAttrib("ToolLevel") > mySettings.getToolLevel()) {
                        aixml::saveDocument(acXML, acXML.at(aircraftXmlParameterPair.first).getIntAttrib("ToolLevel"));
                    } else {
                        aixml::saveDocument(acXML, mySettings.getToolLevel());
                    }
                } else {
                    aixml::saveDocument(acXML, mySettings.getToolLevel());
                }
#endif
            } else {
                std::cout << "Node "<< aircraftXmlParameterPair.first  <<" not found.\n";
                FAIL();
            }
            //aixml::saveDocument(acXML, 3);
            aixml::closeDocument(acXML);
        } else {
            std::cout << "Invalid Input file used. Path of input parameter should begin with either /ConfigFile or /AcftExchangeFile" << std::endl;
            FAIL();
        }
    } catch(const std::invalid_argument& ia) {
        configXML.at(testParameterPath) = testParameterValue;
    } catch (const char * text) {
        std::cout << text << std::endl;
        exit(1);
    } catch (const std::string& text) {
        std::cout << text << std::endl;
        exit(1);
    } catch (const std::exception& e) {
        std::cout << "The following C++ standard error has occured: " << e.what() << std::endl;
        exit(1);
    } catch (...) {
        std::cout << "Invalid Test Parameter input." << std::endl;
        FAIL();
    }

    //Phase 3: Execution of main()
    if (getTestCaseMap()[completeName].executability) {
        executeModule();
    } else {
        if (!aircraftXmlParameterPair.first.empty()) {
            const settings mySettings(this->configXML);
            node& acXML(aixml::openDocument(mySettings.getIoDir() + mySettings.getIoFilename())); // Open aircraft xml file
            if (isNumber(aircraftXmlParameterPair.second)) {
                acXML.at(aircraftXmlParameterPair.first) = std::stod(this->preTestValue); // Reset altered aircraft xml parameter to original value
            } else {
                acXML.at(aircraftXmlParameterPair.first) = this->preTestValue; // Reset altered aircraft xml parameter to original value
            }
#ifdef ENABLE_TOOL_LEVELS
            if (acXML.at(aircraftXmlParameterPair.first).hasAttrib("ToolLevel")) {
                if (acXML.at(aircraftXmlParameterPair.first).getIntAttrib("ToolLevel") > mySettings.getToolLevel()) {
                    aixml::saveDocument(acXML, acXML.at(aircraftXmlParameterPair.first).getIntAttrib("ToolLevel"));
                } else {
                    aixml::saveDocument(acXML, mySettings.getToolLevel());
                }
            } else {
                aixml::saveDocument(acXML, mySettings.getToolLevel());
            }
#endif
            //aixml::saveDocument(acXML, 3); // Save aircraft xml file with toollevel 3 (Otherwise certain values, e.g. MTOM cannot be overwritten)
            aixml::closeDocument(acXML);
            this->aircraftXmlParameterPair.first.clear(); // Clear to be altered pair of aircraft xml parameter + value
            this->aircraftXmlParameterPair.second.clear();
            this->preTestValue.clear();
        }
        GTEST_SKIP();
    }

    //Phase 4: AssertionsF
    applyAssertions(criticalParamsNode, currentTestCase, testParameterPath);
    //Phase 5: Cleanup
    TeardownTest();
}

void BlackboxTest::standardDeathTest() {
    std::string completeName = currentTestSuiteName + "." + currentTestCaseName;
    std::string testParameterPath = getTestParamPathsMap()[completeName.substr(0, completeName.find("("))];

        //Phase 1: Setup
        SetupTest();

        //Phase 2: Alteration of test parameter
        node* currentTestCase = getTestCaseMap()[completeName].testCaseNode;
        std::string testParameterValue = currentTestCase->at("/TestCaseValue");
        try {
            if (testParameterPath.find("ConfigFile") != std::string::npos) {
                if (isNumber(testParameterValue)) {
                    configXML.at(testParameterPath) = std::stod(testParameterValue);
                } else {
                    configXML.at(testParameterPath) = testParameterValue;
                }
                myRuntimeInfo->warn << "Switched parameter at path " << testParameterPath << " to value " << testParameterValue << std::endl;
            } else if (testParameterPath.find("AcftExchangeFile") != std::string::npos) {
                aircraftXmlParameterPair = std::make_pair(testParameterPath, testParameterValue);
                const settings mySettings(this->configXML);
                node& acXML(aixml::openDocument(mySettings.getIoDir() + mySettings.getIoFilename()));
                if (acXML.find(aircraftXmlParameterPair.first)) {
                    const std::string DUMMYpreTestValue = acXML.at(aircraftXmlParameterPair.first);
                    this->preTestValue = DUMMYpreTestValue;
                    if (isNumber(aircraftXmlParameterPair.second)) {
                        acXML.at(aircraftXmlParameterPair.first) = std::stod(aircraftXmlParameterPair.second);
                    } else {
                        acXML.at(aircraftXmlParameterPair.first) = aircraftXmlParameterPair.second;
                    }
                    myRuntimeInfo->warn << "Switched parameter at path " << aircraftXmlParameterPair.first << " to value " << aircraftXmlParameterPair.second << std::endl;
#ifdef ENABLE_TOOL_LEVELS
                    if (acXML.at(aircraftXmlParameterPair.first).hasAttrib("ToolLevel")) {
                        if (acXML.at(aircraftXmlParameterPair.first).getIntAttrib("ToolLevel") > mySettings.getToolLevel()) {
                            aixml::saveDocument(acXML, acXML.at(aircraftXmlParameterPair.first).getIntAttrib("ToolLevel"));
                        } else {
                            aixml::saveDocument(acXML, mySettings.getToolLevel());
                        }
                    } else {
                        aixml::saveDocument(acXML, mySettings.getToolLevel());
                    }
#endif
                } else {
                    std::cout << "Node "<< aircraftXmlParameterPair.first  <<" not found.\n";
                    FAIL();
                }
                //aixml::saveDocument(acXML, 3);
                aixml::closeDocument(acXML);
            } else {
                std::cout << "Invalid Input file used. Path of input parameter should begin with either /ConfigFile or /AcftExchangeFile" << std::endl;
                FAIL();
            }
        } catch(const std::invalid_argument& ia) {
            configXML.at(testParameterPath) = testParameterValue;
        } catch (const char * text) {
            std::cout << text << std::endl;
            exit(1);
        } catch (const std::string& text) {
            std::cout << text << std::endl;
            exit(1);
        } catch (const std::exception& e) {
            std::cout << "The following C++ standard error has occurred: " << e.what() << std::endl;
            exit(1);
        } catch (...) {
            std::cout << "Invalid Test Parameter input." << std::endl;
            FAIL();
            //break;
        }
        //Phase 3/4: Execution of main() as DeathTest
        currentAssertion = DEATH;
#ifdef ENABLE_INDUCED_DEATH_TESTS
        //If Test Suite Name contains "InducedDeathTest" the test should fail if executeModule() exits. The death test has to be inverted
        if (currentTestSuiteName.find("InducedDeathTest") != std::string::npos) {
            EXPECT_EXIT({
                        executeModule();
                        exit(-1);
                        }, ::testing::ExitedWithCode(-1), "") << generateErrorMessage(true, testParameterPath, testParameterValue);
            if (!HasFailure()) {
                std::string correspondingTest = currentTestSuiteName.erase(currentTestSuiteName.size() - 16) +  "." + currentTestCaseName;
                getTestCaseMap()[correspondingTest].executability = true;
            }
        } else {
#endif
            EXPECT_DEATH(executeModule(), "") << generateErrorMessage(false, testParameterPath, testParameterValue);
#ifdef ENABLE_INDUCED_DEATH_TESTS
        }
#endif

        //Phase 5: Cleanup
        TeardownTest();
}

void BlackboxTest::registerAircraft(node* aircraftNode) {
    this->test_acXML = aircraftNode;
}

int BlackboxTest::runAllTests(int argc, char *argv[], std::string tool_name) {
    BlackboxTest::blackboxTestData = &aixml::openDocument(std::string(blackboxTestCasesXmlPath));
    BlackboxTest::registerTests(tool_name);

    testing::GTEST_FLAG(color) = "yes";
    testing::GTEST_FLAG(output) = "xml";
    testing::InitGoogleTest(&argc, argv);

    testing::TestEventListeners& listeners = testing::UnitTest::GetInstance()->listeners();
    testing::TestEventListener* def = listeners.Release(listeners.default_result_printer());
    listeners.Append(new BetterPrinter(def));
    AddGlobalTestEnvironment(new Environment);

    int result = RUN_ALL_TESTS();
    delete def;
    return result;
}

BlackboxTest::BetterPrinter::BetterPrinter(TestEventListener * other): other(other) {}

void BlackboxTest::BetterPrinter::OnTestProgramStart(const testing::UnitTest& unit_test) {
    other->OnTestProgramStart(unit_test);
}

void BlackboxTest::BetterPrinter::OnTestIterationStart(const testing::UnitTest& unit_test, int iteration) {
    other->OnTestIterationStart(unit_test, iteration);
}

void BlackboxTest::BetterPrinter::OnEnvironmentsSetUpStart(const testing::UnitTest& unit_test) { //cppcheck-suppress unusedFunction
    other->OnEnvironmentsSetUpEnd(unit_test);
}
void BlackboxTest::BetterPrinter::OnEnvironmentsSetUpEnd(const testing::UnitTest& unit_test) {
    other->OnEnvironmentsSetUpEnd(unit_test);
}

void BlackboxTest::BetterPrinter::OnTestSuiteStart(const testing::TestSuite& test_suite) {
    other->OnTestSuiteStart(test_suite);
}

void BlackboxTest::BetterPrinter::OnTestStart(const testing::TestInfo& test_info) {
    other->OnTestStart(test_info);
}

void BlackboxTest::BetterPrinter::OnTestPartResult(const testing::TestPartResult& test_part_result) { //cppcheck-suppress unusedFunction
    // http://docs.ros.org/en/jade/api/self_test/html/classtesting_1_1internal_1_1PrettyUnitTestResultPrinter.html#a7589e8df7485349498a3a81bf16e2f68
    if (test_part_result.type() == testing::TestPartResult::kSuccess) return;
    const char* message = test_part_result.message();
    std::string output;
    if (message == nullptr) {
        output += "Unable to provide further information";
        std::cout << output << std::endl;
        return;
    }

    std::string message_s(message);
    output += message_s.erase(0, message_s.find("&&&")+3);
    std::cout << output << std::endl;
}

void BlackboxTest::BetterPrinter::OnTestEnd(const testing::TestInfo& test_info) {
    other->OnTestEnd(test_info);
}

void BlackboxTest::BetterPrinter::OnTestSuiteEnd(const testing::TestSuite& test_suite) {
    other->OnTestSuiteEnd(test_suite);
}

void BlackboxTest::BetterPrinter::OnEnvironmentsTearDownStart(const testing::UnitTest& unit_test) {
    other->OnEnvironmentsTearDownStart(unit_test);
}

void BlackboxTest::BetterPrinter::OnEnvironmentsTearDownEnd(const testing::UnitTest& unit_test) {
    other->OnEnvironmentsTearDownEnd(unit_test);
}

void BlackboxTest::BetterPrinter::OnTestIterationEnd(const testing::UnitTest& unit_test, int iteration) {
    other->OnTestIterationEnd(unit_test, iteration);
}

void BlackboxTest::BetterPrinter::OnTestProgramEnd(const testing::UnitTest& unit_test) {
    other->OnTestProgramEnd(unit_test);
}

bool BlackboxTest::isNumber(const std::string& str) {
    //return str.find_first_not_of("0123456789e+-") == std::string::npos;
    char* end = nullptr;
    double val = strtod(str.c_str(), &end);
    return end != str.c_str() && *end == '\0' && val != std::numeric_limits<double>::max();
}
