/**
 * @file strategySelector.h
 * @author Christopher Ruwisch (christopher.ruwisch@tu-berlin.de)
 * @brief Default strategy selector for basic strategy
 * @version 0.1
 * @date 2023-04-16
 * 
 * @copyright Copyright (c) 2023
 * 
 */
#ifndef MODULEBASICS_INCLUDE_MODULEBASICS_STRATEGYSELECTOR_H_
#define MODULEBASICS_INCLUDE_MODULEBASICS_STRATEGYSELECTOR_H_

#include <moduleBasics/runtimeIO.h>
#include <memory>
#include <utility>

/**
 * \brief Base Strategy class with pure virtual implementation
 * 
 */
class Strategy {
 public:
    /**
     * \brief Initialize your strategy (read your configuration information and data)
     * 
     */
    virtual void initialize() = 0;

    /**
     * \brief Run your strategy (add your algorithms here)
     * 
     */
    virtual void run() = 0;

    /**
     * \brief Update values in aircraft xml via your module data methods (update(...) or update...(...))
     * 
     */
    virtual void update() = 0;

    /**
     * \brief Define and generate plots, define and generate reports
     * 
     */
    virtual void report() = 0;

    /**
     * \brief Save module specific xml's if necessary (do not close acxml or moduleconfig node) from RuntimeIO object
     * 
     */
    virtual void save() = 0;

    /**
     * \brief Destroy the Strategy object
     * 
     */
    virtual ~Strategy() {}
};

/**
 * \brief Strategy selector class for strategies class
 * 
 */
class StrategySelector {
 public:
    /**
     * \brief Set the Strategy object
     * 
     * \param strategy 
     */
    void setStrategy(std::unique_ptr<Strategy> strategy) {
        strategy_ = std::move(strategy);
    }

    /**
     * \brief Run initialize method of strategy
     * 
     */
    void initializeStrategy() {
        strategy_->initialize();
    }

    /**
     * \brief Run run method of strategy
     * 
     */
    void runStrategy() {
        strategy_->run();
    }

    /**
     * \brief Run update method of strategy
     * 
     */
    void updateStrategy() {
        strategy_->update();
    }

    /**
     * \brief Run report method of strategy
     * 
     */
    void reportStrategy() {
        strategy_->report();
    }

    /**
     * \brief Save report method of strategy
     * 
     */
    void saveStrategy() {
        strategy_->save();
    }

 private:
    std::unique_ptr<Strategy> strategy_; /* Selected strategy object by setStrategy */
};

using strategyptr = std::unique_ptr<Strategy>(const std::shared_ptr<RuntimeIO>&);
using strategyaccess = std::function<strategyptr>;

#endif // MODULEBASICS_INCLUDE_MODULEBASICS_STRATEGYSELECTOR_H_
