/*  Copyright (C) 2008-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef STANDARDFILES_INCLUDE_STANDARDFILES_SETTINGS_H_
#define STANDARDFILES_INCLUDE_STANDARDFILES_SETTINGS_H_

#include <string>
#include <cstdint>

#ifdef BUILD_STANDARDFILES_SHARED
    #ifdef _WIN32
        #define STANDARDFILESDLLEXPORT __declspec(dllexport)
    #else
        #define STANDARDFILESDLLEXPORT __attribute__ ((visibility ("default")))
    #endif
#elif defined(IMPORT_STANDARDFILES_SHARED)
    #ifdef _WIN32
        #define STANDARDFILESDLLEXPORT __declspec(dllimport)
    #else
        #define STANDARDFILESDLLEXPORT
    #endif
#else
    #define STANDARDFILESDLLEXPORT
#endif

class node;

/**
*  Global class for program settings that can be accessed within the entire program
**/

class STANDARDFILESDLLEXPORT settings {
 public:
    /** \brief Constructor reads in all tool settings in the configuration file
    *   \details Reads all tool settings and checks paths for project folder, aircraft xml, plot programs (gnuplot and
    *            imagemagick). Furthermore it creates folders for plots, HTML- and tex-reports and creates objects for
    *            the mission specific settings.
    */
    explicit settings(const node& configXML);
    virtual ~settings();


    /** \brief Constant getter function returns the value of the attribute
    *   \return ioFilename std::string-value of aircraft xml filename
    */
    std::string getIoFilename() const;

    /** \brief Constant getter function returns the value of the attribute
    *   \return aeroDataDir std::string-value of aerodynamic data directory
    */
    std::string getAeroDataDir() const;

    /** \brief Constant getter function returns the value of the attribute
    *   \return engineDataDir std::string-value of engine data directory
    */
    std::string getEngineDataDir() const;

    /** \brief Constant getter function returns the value of the attribute
    *   \return geometryDir std::string-value of geometry data directory
    */
    std::string getGeometryDir() const;

    /** \brief Constant getter function returns the value of the attribute
    *   \return airfoilDataDir std::string-value of airfoil data directory
    */
    std::string getAirfoilDataDir() const;

    /** \brief Constant getter function returns the value of the attribute
    *   \return missionDataDir std::string-value of mission data directory
    */
    std::string getMissionDataDir() const;

    /** \brief Constant getter function returns the value of the attribute
    *   \return reportDir std::string-value of report directory
    */
    std::string getReportDir() const;

    /** \brief Constant getter function returns the value of the attribute
    *   \return csvFilesDir std::string-value of csvFilesForPlots directory
    */
    std::string getCsvFilesDir() const;

    /** \brief Constant getter function
     * \return std::string value contains the path of the tools local csv files directory
     */
    std::string getCsvFilesDirTool() const;

    /** \brief Constant getter function
     * \return std::string value containing the path of the plt files directory
     */
    std::string getPltFilesDir() const;

    /** \brief Constant getter function
     * \return std::string value contains the path of the tools local plt files directory
     */
    std::string getPltFilesDirTool() const;

    /** \brief Constant getter function returns the value of the attribute
    *   \return logFilesDir std::string-value of logFiles directory
    */
    std::string getLogFilesDir() const;

    /** \brief Constant getter function returns the value of the attribute
    *   \return ioDir std::string-value of project directory
    */
    std::string getIoDir() const;

    /** \brief Constant getter function returns the value of the attribute
    *   \return plotscript std::string-value of plots script
    */
    std::string getPlotscript() const;

    /** \brief Constant getter function returns the value of the attribute
    *   \return gnuplotPath std::string-value of gnuplot directory
    */
    std::string getGnuplotPath() const;

    /** \brief Constant getter function returns the value of the attribute
    *   \return inkscapePath std::string-value of inkscape directory
    */
    std::string getInkscapePath() const;

    /** \brief Constant getter function returns the value of the attribute
    *   \return toolLevel Integer-value of toolLevel
    */
    int getToolLevel() const;

    /** \brief Constant getter function returns the value of the attribute
    *   \return comments_on uint16_t: console output level 0: no output; 1: only std-out and error; 2: info; 3: debug stored in comments_on
    */
    uint16_t getComments_on() const;

    /** \brief Returns whether log file is written or not
    *   \return logfile_on uint16_t: logfile output level 0: no output; 1: only std-out and error; 2: info; 3: debug stored in logfile_on
    */
    uint16_t getLogfile_on() const;

    /** \brief Constant getter function returns the value of the attribute
    *   \return texReport_on Boolean-value/switch if tex report should be created
    */
    bool getTexReport_on() const;

    /** \brief Constant getter function returns the value of the attribute
    *   \return report_on Boolean-value/switch if a report should be created
    */
    bool getReport_on() const;

    /** \brief Constant getter function returns the value of the attribute
    *   \return plot_on Boolean-value/switch if plots should be created
    */
    bool getPlot_on() const;

    /** \brief Constant getter function
     * \return bool value / switch whether plotting files should be copied to projects folder or not
     */
    bool getCopyPlottingFiles_on() const;

    /** \brief Constant getter function
     * \return bool value / switch whether plotting files should be deleted from tool folder or not
     */
    bool getDeletePlottingFiles_on() const;

    /** \brief Constant getter function returns the value of the attribute
    *   \return writeInfoFile Boolean-value/switch if a info file should be created
    */
    bool getWriteInfoFile() const;

    /** \brief Constant getter function returns the value of the attribute
    *   \return texDir std::string-value of tex report directory
    */
    std::string getTexDir() const;

    /** \brief Constant getter function returns the value of the attribute
    *   \return htmlDir std::string-value of html report directory
    */
    std::string getHtmlDir() const;

    /** \brief Constant getter function returns the value of the attribute
    *   \return plotDir std::string-value of plot report directory
    */
    std::string getPlotDir() const;

 protected:
    const node& configXML;

    /** \brief Writes the document after initialization in constructor
    */
    void writeDocument();

    /* Settings that are fixed for all programs */
    std::string ioFilename;/**< Name of the AcftExchangeFile **/
    std::string ioDir;/**< Directory that contains the AcftExchangeFile **/
    std::string airfoilDataDir;/**< Directory that contains the airfoil geometry data files **/
    std::string geometryDir;/**< Directory that contains the geometry data files and the airfoilDataDir **/
    std::string aeroDataDir;/**< Directory that contains the aerodynamic performance data file xxx_polar.xml **/
    std::string engineDataDir;/**< Directory that contains the engine performance data files **/
    std::string missionDataDir;/**< Directory that contains the mission data input and output files **/
    std::string reportDir;/**< Directory that contains the reporting output files and directories **/
    std::string csvFilesDir;/**< Directory that contains the csvFilesForPlot **/
    std::string csvFilesDirTool;/**< Local tool directory which contains the csv files */
    std::string pltFilesDir;/**< Directory that contains the pltFilesForPlot **/
    std::string pltFilesDirTool;/**< Local tool directory which contains the plt files */
    std::string logFilesDir;/**< Directory that contains the logFiles **/
    std::string plotscript;/**< File name of the gnuplot script file **/
    std::string gnuplotPath;/**< Path to the gnuplot.exe **/
    std::string inkscapePath;/**< Path to the convert.exe of imagemagick **/

    int toolLevel;/**< Tool level of the current program **/
    uint16_t comments_on;/**< Number of console output mode during the program runtime: 0: no output; 1: only std-out and error; 2: info; 3: debug **/
    uint16_t logfile_on;/**< Number of logfile output mode during the program runtime: 0: no output; 1: only std-out and error; 2: info; 3: debug **/
    bool plot_on;/**< Switch for plots **/
    bool copyPlottingFiles_on; /**< Switch for copying plotting files to projects folder **/
    bool deletePlottingFiles_on; /**< Switch for delete plotting files from tool folder**/
    bool report_on;/**< Switch for HTML report **/
    bool texReport_on;/**< Switch for Tex report **/
    bool writeInfoFile;/**< Switch for info file **/

    std::string texDir;/**< Directory that contains the TeX report files **/
    std::string htmlDir;/**< Directory that contains the HTML report files **/
    std::string plotDir;/**< Directory that contains the plot files **/

 private:
    /** \brief Initialize gnuplotPath and inkscapePath
    * \details if in configuration file a path is set to DEFAULT, the default path gnuplot/gnuplot + FILEEXTENSION and inkscape/inkscape + FILEEXTENSION is used
    */
    void initializePlottingPath();

    /** \brief Initialize pathes to subfolders in ioDir
    * \details sets global paths to airfoilDataDir, geometryDir, engineDataDir, aeroDataDir, missionDataDir
    */
    void initializeDirectories();
};

#endif // STANDARDFILES_INCLUDE_STANDARDFILES_SETTINGS_H_
