/*  Copyright (C) 2008-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "standardFiles/settings.h"
#ifdef _WIN32
    #include <windows.h>
#endif // _WIN32
#include <aixml/node.h>
#include <runtimeInfo/runtimeInfo.h>
#include "standardFiles/functions.h"

settings::settings(const node& configXML)
    :
    configXML(configXML),
    ioFilename(configXML.at("IOFileName")),
    ioDir(configXML.at("ControlSettings/IODir")),
    plotscript(configXML.at("ControlSettings/GnuplotScript")),
    gnuplotPath(configXML.at("ControlSettings/GnuplotPath")),
    inkscapePath(configXML.at("ControlSettings/InkscapePath")),
    toolLevel(configXML.at("ControlSettings/OwnToolLevel")),
    comments_on(configXML.at("ControlSettings/ConsoleOutputOn")),
    logfile_on(configXML.at("ControlSettings/LogfileOutputOn")),
    plot_on(configXML.at("ControlSettings/PlotOutputOn")),
    copyPlottingFiles_on(configXML.at("ControlSettings/PlotOutputOn").getBoolAttrib("CopyPlottingFiles")),
    deletePlottingFiles_on(configXML.at("ControlSettings/PlotOutputOn").getBoolAttrib("DeletePlottingFilesFromToolFolder")),
    report_on(configXML.at("ControlSettings/ReportOutputOn")),
    texReport_on(configXML.at("ControlSettings/TexReportOn")),
    writeInfoFile(configXML.at("ControlSettings/WriteInfoFiles")) {
    //ctor fuer Unterklasse
    this->initializePlottingPath();
    this->initializeDirectories();
}

settings::~settings() {
    //dtor
}

void settings::initializePlottingPath() {
    std::string pathForComparison(this->gnuplotPath);
    //GNUPLOT
    for (char& tmpPath : pathForComparison) { //transform gnuplot path to upper case values
        tmpPath = ::toupper(tmpPath);
    }
    if (compareStrings(pathForComparison, "DEFAULT", true)) {
        this->gnuplotPath = GNUPLOTDEFAULTDIR;
    }
    //INKSCAPE
    pathForComparison = this->inkscapePath;
    for (char& tmpPath : pathForComparison) { //transform inkscape path to upper case values
        tmpPath = ::toupper(tmpPath);
    }
    if (compareStrings(pathForComparison, "DEFAULT", true)) {
        this->inkscapePath = INKSCAPEDEFAULTDIR;
    }
}

void settings::initializeDirectories() {
    this->geometryDir = this->getIoDir() + "geometryData" + FILESEPERATOR;
    this->airfoilDataDir = this->getGeometryDir() + "airfoilData" + FILESEPERATOR;
    this->aeroDataDir = this->getIoDir() + "aeroData" + FILESEPERATOR;
    this->engineDataDir = this->getIoDir() + "engineData" + FILESEPERATOR;
    this->missionDataDir = this->getIoDir() + "missionData" + FILESEPERATOR;
    this->reportDir = this->getIoDir() + "reporting" + FILESEPERATOR;
    this->texDir = this->reportDir + "reportTeX" + FILESEPERATOR;
    this->htmlDir = this->reportDir + "reportHTML" + FILESEPERATOR;
    this->plotDir = this->reportDir + "plots" + FILESEPERATOR;
    this->csvFilesDir = this->plotDir + "csvFiles" + FILESEPERATOR;
    this->csvFilesDirTool = std::string("csvFiles") + FILESEPERATOR;
    this->pltFilesDir = this->plotDir + "pltFiles" + FILESEPERATOR;
    this->pltFilesDirTool = std::string("pltFiles") + FILESEPERATOR;
    this->logFilesDir = this->getReportDir() + "logFiles" + FILESEPERATOR;
}

/* Getter Functions */
std::string settings::getIoFilename() const {
    return this->ioFilename;
}

std::string settings::getIoDir() const {
    #ifdef _WIN32
        return replaceAll(this->ioDir, "/", FILESEPERATOR);
    #else
        return win2lin(this->ioDir);
    #endif // _WIN32
}

std::string settings::getAeroDataDir() const { // cppcheck-suppress unusedFunction
    return this->aeroDataDir;
}

std::string settings::getAirfoilDataDir() const { // cppcheck-suppress unusedFunction
    return this->airfoilDataDir;
}

std::string settings::getEngineDataDir() const { // cppcheck-suppress unusedFunction
    return this->engineDataDir;
}

std::string settings::getMissionDataDir() const { // cppcheck-suppress unusedFunction
    return this->missionDataDir;
}

std::string settings::getGeometryDir() const {
    return this->geometryDir;
}

std::string settings::getReportDir() const {
    return this->reportDir;
}

std::string settings::getCsvFilesDir() const { // cppcheck-suppress unusedFunction
    return this->csvFilesDir;
}

std::string settings::getCsvFilesDirTool() const { // cppcheck-suppress unusedFunction
    return this->csvFilesDirTool;
}

std::string settings::getPltFilesDir() const { // cppcheck-suppress unusedFunction
    return this->pltFilesDir;
}

std::string settings::getPltFilesDirTool() const { // cppcheck-suppress unusedFunction
    return this->pltFilesDirTool;
}

std::string settings::getLogFilesDir() const { // cppcheck-suppress unusedFunction
    return this->logFilesDir;
}

std::string settings::getPlotscript() const { // cppcheck-suppress unusedFunction
    return this->plotscript;
}

std::string settings::getGnuplotPath() const { // cppcheck-suppress unusedFunction
    return this->gnuplotPath;
}

std::string settings::getInkscapePath() const { // cppcheck-suppress unusedFunction
    return this->inkscapePath;
}

int settings::getToolLevel() const { // cppcheck-suppress unusedFunction
    return this->toolLevel;
}

uint16_t settings::getComments_on() const { // cppcheck-suppress unusedFunction
    return this->comments_on;
}

uint16_t settings::getLogfile_on() const { // cppcheck-suppress unusedFunction
    return this->logfile_on;
}

bool settings::getTexReport_on() const { // cppcheck-suppress unusedFunction
    return this->texReport_on;
}

bool settings::getReport_on() const { // cppcheck-suppress unusedFunction
    return this->report_on;
}

bool settings::getPlot_on() const { // cppcheck-suppress unusedFunction
    return this->plot_on;
}

bool settings::getCopyPlottingFiles_on() const { // cppcheck-suppress unusedFunction
    return this->copyPlottingFiles_on;
}

bool settings::getDeletePlottingFiles_on() const { // cppcheck-suppress unusedFunction
    return this->deletePlottingFiles_on;
}

bool settings::getWriteInfoFile() const { // cppcheck-suppress unusedFunction
    return this->writeInfoFile;
}

std::string settings::getTexDir() const { // cppcheck-suppress unusedFunction
    return this->texDir;
}

std::string settings::getHtmlDir() const { // cppcheck-suppress unusedFunction
    return this->htmlDir;
}

std::string settings::getPlotDir() const { // cppcheck-suppress unusedFunction
    return this->plotDir;
}

void settings::writeDocument() { // cppcheck-suppress unusedFunction
    // Check paths for project folder and Aircraft-XML
    if (!fileExists(this->getIoDir())) {
        myRuntimeInfo->err << "The project folder " << ioDir << ", that is given in the config file, could not be found. Abort program." << std::endl;
        exit(1);
    } else {
        if (!fileExists(this->getIoDir() + this->getIoFilename())) {
            myRuntimeInfo->err << "The Aircraft-Exchange-File (" << this->getIoFilename();
            myRuntimeInfo->err << "), that is given in the config file, could not be found in the given project folder (" << this->getIoDir() << "). Abort program." << std::endl;
            exit(1);
        }
    }

    if (!fileExists(this->reportDir) && (this->texReport_on || this->report_on || this->plot_on || this->logfile_on)) {
        myRuntimeInfo->out << "Create folder for reporting in project folder: " << reportDir << std::endl;
        CreateDirectory(reportDir);
    }
    if (!fileExists(texDir) && texReport_on) {
        myRuntimeInfo->out << "Create folder for tex-report in project folder: " << texDir << std::endl;
        CreateDirectory(texDir);
    }
    if (!fileExists(htmlDir) && report_on) {
        myRuntimeInfo->out << "Create folder for html-report in project folder: " << htmlDir << std::endl;
        CreateDirectory(htmlDir);
    }
    if (!fileExists(plotDir) && plot_on) {
        myRuntimeInfo->out << "Create folder for plots in project folder: " << plotDir << std::endl;
        CreateDirectory(plotDir);
    }
    if (!fileExists(csvFilesDir) && plot_on) {
        CreateDirectory(csvFilesDir);
    }
    if (!fileExists(pltFilesDir) && plot_on) {
        CreateDirectory(pltFilesDir);
    }
    /* Check path for plotting programs */
    if (!fileExists(gnuplotPath) && plot_on) {
        std::stringstream errorMsg;
        errorMsg << "Gnuplot-Path \"" << gnuplotPath << "\" could not be found. Change settings in config file. Abort program.";
        throwError(__FILE__, __func__, __LINE__, errorMsg.str());
    }
    if (!fileExists(inkscapePath) && plot_on && texReport_on) {
        std::stringstream errorMsg;
        errorMsg << "Inkscape-Path \"" << inkscapePath << "\" could not be found. Change settings in config file. Abort program.";
        throwError(__FILE__, __func__, __LINE__, errorMsg.str());
    }
}
